#' 
#' Poisson kernel-based quadratic distance test of Uniformity on the sphere
#' 
#' This function performs the kernel-based quadratic distance goodness-of-fit tests for Uniformity for spherical data using the Poisson kernel with concentration parameter \code{rho}.
#'
#' @param x A numeric d-dim matrix of data points on the Sphere S^(d-1).
#' @param rho Concentration parameter of the Poisson kernel function.
#' @param B Number of iterations for critical value estimation of Un (default: 300).
#' @param Quantile The quantile to use for critical value estimation, 0.95 is the default value.
#'
#' @return An S4 object of class \code{pk.test} containing the results of the Poisson kernel-based tests. The object contains the following slots:
#'\itemize{
#'   \item \code{method}: String indicating that the Poisson Kernel-based test is performed.
#'   \item \code{x} Data matrix.
#'   \item \code{Un} The value of the U-statistic.
#'   \item \code{CV_Un} The empirical critical value for Un.
#'   \item \code{H0_Vn} A logical value indicating whether or not the null hypothesis is rejected according to Un.
#'   \item \code{Vn} The value of the V-statistic.
#'   \item \code{CV_Vn} The critical value for Vn computed following the asymptotic distribution.
#'   \item \code{H0_Vn} A logical value indicating whether or not the null hypothesis is rejected according to Vn.
#'   \item \code{rho} The value of concentration parameter used for the Poisson kernel function.
#'   \item \code{B} Number of replications for the critical value of the U-statistic.
#'}
#'
#'
#' @references
#' Ding, Y., Markatou, M., Saraceno, G. (2023). “Poisson Kernel-Based Tests for
#' Uniformity on the d-Dimensional Sphere.” Statistica Sinica. 
#' doi: doi:10.5705/ss.202022.0347
#' 
#' @examples
#' # create a pk.test object
#' x_sp <- sample_hypersphere(3, n_points=100)
#' unif_test <- pk.test(x_sp,rho=0.8)
#' unif_test
#'
#' @importFrom stats qchisq
#'
#' @useDynLib QuadratiK
#' @importFrom Rcpp sourceCpp
#'
#' @export
setGeneric("pk.test",function(x, rho = NULL, B = 300, Quantile = 0.95){
   standardGeneric("pk.test")})

#' @rdname pk.test
#' @export
setMethod("pk.test", signature(x = "ANY"),
          function(x,  rho, B = 300, Quantile = 0.95)
          {
             
             # Convert vectors to a single column matrix
             if(is.vector(x)) {
                x <- matrix(x, ncol = 1)
             } else if(is.data.frame(x)) {
                x <- as.matrix(x)
             } else if(!is.matrix(x)){
                stop("x must be a matrix or a data.frame")
             }
             
             if(Quantile<=0 | Quantile>1){
                stop("Quantile indicates the level used for the critical value computation. It must be in (0,1].")
             }
             if(!is.numeric(rho) | (rho<=0 | rho>1)){
                stop("rho indicates the concentration parameter of the Poisson kernel, it must be in (0,1).")
             }
             
             
             METHOD <- "Poisson Kernel-based quadratic distance test of Uniformity on the Sphere"
             
             d<- ncol(x)
             n <- nrow(x)
             
             pk <- statPoissonUnif(x, rho)
             
             #x.prod <- x%*%t(x)
             #pk <- ((1-rho^2)/ (( 1 + rho^2 - 2*rho*x.prod)^(d/2))) - 1
             
             dof <- DOF(d, rho)
             qu_q <- qchisq(Quantile,df=dof$DOF)
             CV_Vn <- dof$Coefficient*qu_q
             
             var_Un <- (2/(n*(n-1)))*((1+rho^2)/((1-rho^2)^(d-1)) -1)
             
             CV_Un <- poisson_CV(d=d, size=n, rho=rho, B=B, Quantile=Quantile )
             
             res <- new("pk.test", Un = pk[1]/sqrt(var_Un), CV_Un = CV_Un, Vn = pk[2], CV_Vn = CV_Vn, method = METHOD, x = x, B= B, rho= rho, H0_Un = pk[1]/sqrt(var_Un) > CV_Un, H0_Vn = pk[2] > CV_Vn)
             return(res)
          })
#' @rdname pk.test
#'
#' @param object Object of class \code{pk.test}
#'
#' @export
setMethod("show", "pk.test",
          function(object) {
             cat( "\n", object@method, "\n")
             cat("Selected consentration parameter rho: ", object@rho, "\n")
             
             cat("\n")
             cat("U-statistic:\n")
             cat("\n")
             cat("H0 is rejected: ", object@H0_Un , "\n")
             cat("Statistic Un: ", object@Un, "\n")
             cat("Critical value: ", object@CV_Un,"\n")
             
             cat("\n")
             cat("V-statistic:\n")
             cat("\n")
             cat("H0 is rejected: ", object@H0_Vn, "\n")
             cat("Statistic Vn: ", object@Vn, "\n")
             cat("Critical value: ", object@CV_Vn,"\n")
             
             cat("\n")
          })
#' Summarizing kernel-based quadratic distance results
#'
#' \code{summary} method for the class \code{pk.test}
#'
#' @param object Object of class \code{pk.test}
#' 
#' @return List with the following components:
#' \itemize{
#'    \item \code{summary_tables} Table of computed descriptive statistics per variable.
#'    \item \code{test_results} Data frame with the results of the performed Poisson kernel-based test.
#'    \item \code{qqplots} Figure with qq-plots for each variable against the uniform distribution.
#' }
#'
#' @import ggpubr
#' @import ggplot2
#' 
#' @examples
#' # create a pk.test object
#' x_sp <- sample_hypersphere(3, n_points=100)
#' unif_test <- pk.test(x_sp,rho=0.8)
#' summary(unif_test)
#'
#' @export
#'
setMethod("summary", "pk.test", function(object) {
   
   dat_x <- as.data.frame(object@x)
   
   figure <- NA
   
   plot_list <- list()
   stats <- list()
   for(i in 1:ncol(dat_x)) {
      
      unif_data <- runif(nrow(dat_x),-1,1)
      qq_df <- data.frame(x = quantile(unif_data, probs = seq(0, 1, length.out = nrow(dat_x))), sample_quantiles = quantile(dat_x[,i], probs = seq(0, 1, length.out = nrow(dat_x))))
      
      pl <- ggplot(qq_df, aes(x = qq_df$x, y = qq_df$sample_quantiles)) +
         geom_line(col="blue") +
         theme_minimal()+
         geom_abline(slope = 1, intercept = 0,col="red") +
         ggtitle(paste("QQ Plot against Uniform - ",names(dat_x)[i])) +
         xlab("Theoretical Quantiles") +
         ylab("Sample Quantiles")
      
      
      stats_step <- data.frame(matrix(c(mean(dat_x[,i]),sd(dat_x[,i]),median(dat_x[,i]),IQR(dat_x[,i]),min(dat_x[,i]),max(dat_x[,i])),nrow=6,ncol=1,byrow=TRUE))
      colnames(stats_step) <- c(paste(names(dat_x)[i]))
      rownames(stats_step) <- c("mean", "sd", "median", "IQR", "min", "max")
      
      stats[[i]] <- stats_step
      
      pl_stat <- ggplot() +
         ggpp::annotate('table', x = 0.5, y = 0.5, label = data.frame(Stat = rownames(stats_step),stats_step), hjust = 0.5, vjust = 0.5) +
         theme_void() +
         ggtitle("")+
         scale_color_brewer(palette='Set1')
      
      plot_list[[length(plot_list) + 1]] <- list(pl,pl_stat)
      
      
   }
   stats <- do.call(cbind, stats)
   colnames(stats) <- c(names(dat_x))
   rownames(stats) <- c("mean", "sd", "median", "IQR", "min", "max")
   
   plot_list <- do.call(c, plot_list)
   figure <- ggarrange(plotlist = plot_list, nrow = length(plot_list)/2, ncol = 2)
   
   # Print main results of the test
   cat( "\n", object@method, "\n")
   test_results <- data.frame(
      Test_Statistics = c(object@Un, object@Vn),
      Critical_Value = c(object@CV_Un,object@CV_Vn),
      Reject_H0 = c(object@H0_Un, object@H0_Vn)
   )
   print(test_results)
   print(figure)
   return(list(summary_tables = stats, test_results = test_results, qqplots = figure))
})
